#pragma once

#pragma comment(linker, "/NODEFAULTLIB") 
#pragma comment( lib, "kernel32" )
#pragma comment( lib, "gdi32" )  
#pragma comment(linker, "/ENTRY:\"DllEntryPoint\"") // define entry point cause no C Lib

#define WIN32_LEAN_AND_MEAN
#define INITGUID
#include <windows.h>
#include <guiddef.h>
#include <ddraw.h>
#include <dvp.h>
#include "ddrawi.h"
#include "my_no_crt.h"


#define DEBUG 0

#define SLEEP_INTERVAL 14 // ms between frames for Config.FakeVsync
#define MAX_BACKBUFFERS 2 // of the primary for Config.NoVidMem

struct WRAP
{
    void* vtable;
	long  magic;
	union
	{
		IUnknown*     iunknown;
		IClassFactory* classfactory;
		IDirectDraw*  dd1;
		IDirectDraw2* dd2;
		IDirectDraw4* dd4;
		IDirectDraw7* dd7;
		IDirectDrawSurface*  dds1;
		IDirectDrawSurface2* dds2;
		IDirectDrawSurface3* dds3;
		IDirectDrawSurface4* dds4;
		IDirectDrawSurface7* dds7;
		IDirectDrawClipper*  clipper;
	};
	GUID  iid; 
	volatile long ref;
	WRAP* next;
	WRAP* prev;
};

struct EnumStruct
{
	LPVOID lpOriginalCallback;
	LPVOID lpOriginalContext;
};


// cfg.cpp
struct CONFIG
{
	DWORD Affinity;
	DWORD BltMirror;
	DWORD BltNoTearing;
	DWORD ColorFix;
	DWORD DpiAware;
	DWORD ForceBltNoTearing;
	DWORD ForceHEL;
	DWORD FakeVsync;
	DWORD NoVidMem;
	DWORD Wine_Diablo;
	VOID Init();
	CHAR szDDrawPath[MAX_PATH];
};
BOOL TestVsync( WRAP* dd );
DWORD TestBltMirror( WRAP* dd, DWORD dwDescSize );
DWORD TestFlip( WRAP* dd, DWORD dwDescSize );
extern CONFIG Config;


// helpers.cpp
HRESULT FakeVsync( HANDLE hEvent );
HRESULT GetDDInterface_Internal( WRAP* wrap, LPVOID* lplpDD );
HRESULT CreateSurfaceProc( WRAP* wrap, LPDDSURFACEDESC lpDDSurfaceDesc, LPDIRECTDRAWSURFACE *lplpDDSurface, IUnknown *pUnkOuter) ;
bool RopUsesPattern( DWORD dwRop );
extern HANDLE g_hThread;
extern BOOL g_ShowDisplay;
extern DWORD g_dwTimerId;
extern IDirectDrawSurface* g_lpDDSRealPrimary;
extern IDirectDrawSurface* g_lpDDSFakePrimary[ 1 + MAX_BACKBUFFERS ];

//clipper.cpp
struct IDirectDrawClipper_vtable;
extern IDirectDrawClipper_vtable clip_vtable;

// dd.cpp
struct IDirectDraw1_vtable;
struct IDirectDraw2_vtable;
struct IDirectDraw4_vtable;
struct IDirectDraw7_vtable;
extern IDirectDraw1_vtable dd1_vtable;
extern IDirectDraw2_vtable dd2_vtable;
extern IDirectDraw4_vtable dd4_vtable;
extern IDirectDraw7_vtable dd7_vtable;

// dds.cpp
struct IDirectDrawSurface_vtable;
extern IDirectDrawSurface_vtable dds_vtable;

// unknwn.cpp
struct IUnknown_vtable;
struct IClassFactory_vtable;
extern IUnknown_vtable unkn_vtable;
extern IClassFactory_vtable cf_vtable;

// dllmain.cpp
extern HMODULE g_hDll;

// wrapper.cpp
extern CRITICAL_SECTION WrapCriticalSection;
void DeleteWrap( WRAP* wrapper );
ULONG AddRefWrap( WRAP* wrapper );
ULONG ReleaseWrap( WRAP* wrapper );
bool GetWrap( void** ppv, const GUID& riid );
IUnknown*            GetInterfacePtr( IUnknown*            wrapper );
IDirectDraw*         GetInterfacePtr( IDirectDraw*         wrapper );
IDirectDraw2*        GetInterfacePtr( IDirectDraw2*        wrapper );
IDirectDraw4*        GetInterfacePtr( IDirectDraw4*        wrapper );
IDirectDraw7*        GetInterfacePtr( IDirectDraw7*        wrapper );
IDirectDrawSurface*  GetInterfacePtr( IDirectDrawSurface*  wrapper );
IDirectDrawSurface2* GetInterfacePtr( IDirectDrawSurface2* wrapper );
IDirectDrawSurface3* GetInterfacePtr( IDirectDrawSurface3* wrapper );
IDirectDrawSurface4* GetInterfacePtr( IDirectDrawSurface4* wrapper );
IDirectDrawSurface7* GetInterfacePtr( IDirectDrawSurface7* wrapper );
IDirectDrawClipper*  GetInterfacePtr( IDirectDrawClipper*  wrapper );

HRESULT __stdcall EnumSurfaces1Callback( LPDIRECTDRAWSURFACE  lpDDSurface, LPDDSURFACEDESC  lpDDSurfaceDesc, LPVOID lpContext );
HRESULT __stdcall EnumSurfaces4Callback( LPDIRECTDRAWSURFACE4 lpDDSurface, LPDDSURFACEDESC2 lpDDSurfaceDesc, LPVOID lpContext );
HRESULT __stdcall EnumSurfaces7Callback( LPDIRECTDRAWSURFACE7 lpDDSurface, LPDDSURFACEDESC2 lpDDSurfaceDesc, LPVOID lpContext );

// log.cpp
#if DEBUG
void __cdecl Log( char* fmt, ... );
void LogGUID( const GUID& riid );
#define LOG Log
#define LOGGUID LogGUID
extern "C" void * _ReturnAddress(void);
#pragma intrinsic(_ReturnAddress)
#define RET _ReturnAddress()
#else
#define LOG(...)
#define LOGGUID(x)
#define RET
#endif
