#include "header.h"

IDirectDrawSurface* g_lpDDSRealPrimary = NULL; // There can be only one.
IDirectDrawSurface* g_lpDDSFakePrimary[ 1 + MAX_BACKBUFFERS ] = { 0 };
HANDLE g_hThread;
BOOL g_ShowDisplay;
DWORD g_dwTimerId;

// Config.NoVidMem todo:
//		add checks in dd->EnumSurfaces and dds->EnumAttachedSurfaces
//		clipper/windowed mode?
//		SetSurfaceDescription() / GetSurfaceDescription()?
   
// correct? with rops and/or ddrops
bool RopUsesPattern( DWORD dwRop )
{ 
	return ( (((dwRop & 0x00F00000) >> 4) != (dwRop & 0x000F0000)) || (((dwRop & 0xF0000000) >> 4) != (dwRop & 0x0F000000)) ); 
}

// ColorFix / Stolen Palette fix sets flag 0x800 on a internal ddraw struct member called dwAppHackFlags
//  ( see DirectDrawCreate[Ex] or dd(1...7)->Initialize() )
// if ( flag & 0x10 ) do not disable the screen saver for exclusive mode ??
// else disable by calling SystemParametersInfo( SPI_SCREENSAVERRUNNING, 1 ...) 
// note: SystemParametersInfo( SPI_SCREENSAVERRUNNING, ...) disables alt+tab and ctrl+alt+del on win9x

/*
todo:
1a. HEL and "Windows 7 compatible applications must not lock the primary video buffer to render." doing so results in an error...
	- this is a big deal because HEL is now preferred over HAL... by me.

1b. "Outside of the lock/unlock pair, the surface structure is not guaranteed to reside at the same location, nor even to exist." ( Since Windows 2000 )
	- programs shouldn't be doing this anyways but doesn't mean that they don't.... ( *cough* "Axis & Allies: Iron Blitz" *cough* )
	- this behavior breaks Wine as well AFAIK

2. "Applications manifested for Windows 7 are prevented from performing Blt's to the primary Desktop video buffer without a clipping window."
	- who cares; don't manifested your directdraw app for win7... however the DWM Composition stays on if you do
	- BltFast doesn't work with a clipper, so redirect it thru Blt
	- grab the hWnd from SetCooperativeLevel
	
*/


// interesting discussion on timers: http://forum.sysinternals.com/bug-in-waitable-timers_topic16229.html

// HEL doesn't support (or fake) vsync 
// this is a problem because some animation is frame based and can run too fast
// we can get the real verticle blank from D3DKMT but the screen still tears in HEL though I don't know why???
// todo: what is GetVerticalBlankStatus returning in HEL?
// todo: DDOVER_AUTOFLIP ??
HRESULT FakeVsync( HANDLE hEvent )
{
	static DWORD dwPrevFrameTickCount = 0;
	DWORD dwTickCount = GetTickCount();
	LONG lDiff = SLEEP_INTERVAL - (dwTickCount - dwPrevFrameTickCount);
	if( lDiff > 0 )
	{
		Sleep( lDiff );
		dwTickCount += lDiff;
	}
	dwPrevFrameTickCount = dwTickCount;
	if( hEvent ) SetEvent( hEvent );
	return DD_OK;
} 


                                           


// copy the fake primary to the real primary every so often
//
// WaitForSingleObject( hReleasePrimaryEvent, 14 ) // always enters ring 0... so overkill?
// SetWaitableTimer()? 
//
// this seem too slow: 
//		WaitForVerticalBlankProc( g_lpDDWrap, DDWAITVB_BLOCKBEGIN, 0 );
//		Blt()
// 		Sleep(0);
DWORD __stdcall ThreadProc( VOID* param )
{
	Sleep(300); // so we don't have the surface "busy" while the game tries to init it
	while( g_ShowDisplay )
	{
		HRESULT hResult = g_lpDDSRealPrimary->Blt( NULL, g_lpDDSFakePrimary[0], NULL, DDBLT_WAIT, NULL );
		if( hResult == DDERR_SURFACELOST ) g_lpDDSRealPrimary->Restore();
		Sleep( SLEEP_INTERVAL );  
	}
	return 0;
}


// copy the fake primary to the real primary every so often
//
// AFAIU there should never be more than one WM_TIMER messages in the queue for a normal message loop
VOID __stdcall TimerProc( HWND hWnd, UINT uMsg, UINT idEvent, DWORD dwTickCount )
{
	if( g_ShowDisplay )
	{
		HRESULT hResult = g_lpDDSRealPrimary->Blt( NULL, g_lpDDSFakePrimary[0], NULL, DDBLT_WAIT, NULL ); 
		if( hResult == DDERR_SURFACELOST ) g_lpDDSRealPrimary->Restore();
	}
}


// works for all ddraw surface interface versions
// found DD interface isn't wrapped
HRESULT GetDDInterface_Internal( WRAP* wrap, LPVOID* lplpDD ) 
{ 
	HRESULT hResult; 
	if( wrap->iid == IID_IDirectDrawSurface )
	{
		IDirectDrawSurface2* dds2;
		if( SUCCEEDED( wrap->dds1->QueryInterface( IID_IDirectDrawSurface2, (LPVOID*) &dds2 ) ) )
		{
			hResult = dds2->GetDDInterface( lplpDD );
			dds2->Release();
		}
	}
	else
	{
		hResult = wrap->dds2->GetDDInterface( lplpDD );
	}
	return hResult; 
}


// store the handle to the real primary surface
// if Config.NoVidMem then create a fake primary surface and force offscreen surfaces into system memory 
HRESULT CreateSurfaceProc( WRAP* wrap, LPDDSURFACEDESC lpDDSurfaceDesc, LPDIRECTDRAWSURFACE *lplpDDSurface, IUnknown *pUnkOuter) 
{
	HRESULT hResult;
	if( lpDDSurfaceDesc->dwFlags & DDSD_CAPS && lpDDSurfaceDesc->ddsCaps.dwCaps & DDSCAPS_PRIMARYSURFACE ) 
	{
		hResult = wrap->dd1->CreateSurface( lpDDSurfaceDesc, &g_lpDDSRealPrimary, NULL );
		if( FAILED( hResult ) ) return hResult;

		if( Config.NoVidMem ) 
		{ 
			//// Create Fake Primary Surface ////
			if( SUCCEEDED( g_lpDDSRealPrimary->GetSurfaceDesc( lpDDSurfaceDesc ) ) ) 
			{
				DDSURFACEDESC2 ddsd; 
				memcpy( &ddsd, lpDDSurfaceDesc, lpDDSurfaceDesc->dwSize );
				ddsd.dwFlags &= ~DDSD_BACKBUFFERCOUNT;
				ddsd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;

				// caps not allowed (picked at random)
				ddsd.ddsCaps.dwCaps &= ~( DDSCAPS_OPTIMIZED | DDSCAPS_NONLOCALVIDMEM | DDSCAPS_LOCALVIDMEM | 
					DDSCAPS_ALLOCONLOAD | DDSCAPS_VISIBLE | DDSCAPS_VIDEOMEMORY | DDSCAPS_PRIMARYSURFACE |
					DDSCAPS_FRONTBUFFER | DDSCAPS_FLIP | DDSCAPS_COMPLEX | DDSCAPS_BACKBUFFER | DDSCAPS_OVERLAY );

				if( SUCCEEDED( wrap->dd1->CreateSurface( (LPDDSURFACEDESC)&ddsd, &g_lpDDSFakePrimary[0], pUnkOuter ) ) )
				{
					if( lpDDSurfaceDesc->dwFlags & DDSD_BACKBUFFERCOUNT )
					{
						// create back buffers
						DWORD i = 0;
						for( ; i < lpDDSurfaceDesc->dwBackBufferCount && i < MAX_BACKBUFFERS; i++ ) 
						{
							if( SUCCEEDED( wrap->dd1->CreateSurface( (LPDDSURFACEDESC)&ddsd, &g_lpDDSFakePrimary[ 1 + i ], pUnkOuter ) ) )
							{
								if( SUCCEEDED( g_lpDDSFakePrimary[i]->AddAttachedSurface( g_lpDDSFakePrimary[ 1 + i ] ) ) )
								{
									;
								}
								else break;
							}
							else break;
						}
						if( i < lpDDSurfaceDesc->dwBackBufferCount ) // if error creating back buffers
						{
							for(int ii = MAX_BACKBUFFERS; ii >= 0; ii-- ) // destroy fake primary
							{
								if( g_lpDDSFakePrimary[ii] != NULL )
								{
									g_lpDDSFakePrimary[ii]->DeleteAttachedSurface( 0, NULL );
									g_lpDDSFakePrimary[ii]->Release();	
									g_lpDDSFakePrimary[ii] = NULL;
								}
							}
							*lplpDDSurface = g_lpDDSRealPrimary; // return real primary
							return DD_OK;
						}
					}

					g_lpDDSFakePrimary[0]->AddRef();
					*lplpDDSurface = g_lpDDSFakePrimary[0];

					if(Config.NoVidMem == 2 )
					{
						g_ShowDisplay = TRUE;
						g_dwTimerId = SetTimer( NULL, 0, SLEEP_INTERVAL, TimerProc );	
						//todo: if( g_Timer == 0 ) error();
					}
					if( Config.NoVidMem == 1 )
					{											
						DWORD dwThreadId;
						g_ShowDisplay = TRUE;
						g_hThread = CreateThread( NULL, 0, &ThreadProc, 0, 0, (LPDWORD)&dwThreadId );
						//todo: if( g_hThread == NULL ) error();
						
						SetThreadPriority( g_hThread, GetThreadPriority( GetCurrentThread() ) );
					}
					return DD_OK;	
				}
			} // end GetSurfaceDesc
		} // end NoVidMem
		*lplpDDSurface = g_lpDDSRealPrimary;
		return hResult;
	}  

	//// Create Non-Primary Surface ////
	// put all surfaces in system memory ( if possible )
	if( !( lpDDSurfaceDesc->ddsCaps.dwCaps & DDSCAPS_OVERLAY ) ) // overlays must be in video memory
	{
		DWORD dwOldCaps = lpDDSurfaceDesc->ddsCaps.dwCaps;
		lpDDSurfaceDesc->ddsCaps.dwCaps &= ~(DDSCAPS_VIDEOMEMORY | DDSCAPS_LOCALVIDMEM | DDSCAPS_NONLOCALVIDMEM);
		lpDDSurfaceDesc->ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
		hResult = wrap->dd1->CreateSurface( lpDDSurfaceDesc, lplpDDSurface, pUnkOuter );
		lpDDSurfaceDesc->ddsCaps.dwCaps = dwOldCaps;
		if( SUCCEEDED( hResult ) ) return hResult;
		// else fail thru and try with old caps
	}
	return wrap->dd1->CreateSurface( lpDDSurfaceDesc, lplpDDSurface, pUnkOuter );
}

